indexing
	description: "Objects that represents a date in the genealogical database"
	author: "Thomas Gubler and Wolfgang Schwaiger"
	date: "$Date: 2008-05-20 21:01:05 +0200 (Tue, 20 May 2008) $"
	revision: "$Revision: 14 $"

class
	GEN_DATE

create
	make_from_string, make_nulldate

feature -- Initialisation

	make_from_string (a_date_string: STRING) is
			-- Initializes the date from a string
			require
				string_not_void: a_date_string /= void
				string_length_correct: a_date_string.count = 10
			do
				set_date (a_date_string)
			end

	make_nulldate is
			-- Initializes an undefined Date (null date)
			do
				m_date := void
			end


feature{NONE} -- Implementation

	m_date: STRING

	day_string: STRING is
			-- Gets the day string
			require not_null_date: not is_null_date
			do
				Result:= m_date.substring (1, 2)
			end

	month_string: STRING is
			-- Gets the day string
			require not_null_date: not is_null_date
			do
				Result:= m_date.substring (4, 5)
			end

	year_string: STRING is
			-- Gets the day string
			require not_null_date: not is_null_date
			do
				Result:= m_date.substring (7, 10)
			end

feature {GEN_DATE} -- Date details

	day: INTEGER is
			-- gets the day number
			require
				day_is_defined: is_day_defined
			do
				Result:=day_string.to_integer
			end

	month: INTEGER is
			-- gets the month number
			require
				month_is_defined: is_month_defined
			do
				Result:=month_string.to_integer
			end

	year: INTEGER is
			-- gets the year number
			require
				year_is_defined: is_year_defined
			do
				Result:=year_string.to_integer
			end


	is_day_defined: BOOLEAN is
			-- true if the day is defined, false if the day is undefined ("??")
			require
				not_null_date: not is_null_date
			do
				Result:=day_string.is_integer
			end

	is_month_defined: BOOLEAN is
			-- true if the month is defined, false if the month is undefined ("??")
			require
				not_null_date: not is_null_date
			do
				Result:= month_string.is_integer
			end

	is_year_defined: BOOLEAN is
			-- true if the year is defined, false if the year is undefined ("????")
			require
				not_null_date: not is_null_date
			do
				Result:= year_string.is_integer
			end

	is_full_defined: BOOLEAN is
			-- true if the date is full qualifide (day, month, year), false if at leas one part is undefined ("??")
			require
				not_null_date: not is_null_date
			do
				Result:= (is_day_defined and is_month_defined and is_year_defined)
			end


feature -- Access

	date_string: STRING is
			-- gets the date string
			require
				date_not_null: not is_null_date
			do
				Result := m_date
			end

feature -- Status Report


	is_null_date : BOOLEAN is
			-- true if this instance represents an undifined date
			do
				Result:= m_date = void
			end

	is_valid : BOOLEAN is
			-- true if the date is valid, otherwise false
			require
				not_null: not is_null_date
			local
				date: DATE
			do
				Result := true

				-- Check the year
				if Result and is_year_defined then
					Result := year >= 0
				end

				-- Check Month
				if Result and is_month_defined then
					Result := (month > 0 and month <= 12)
				end

				-- Check Days
				if Result and is_day_defined then
					create{DATE} date.make (year, month,1)
					Result := (day > 0 and day <= date.days_in_month)
				end
			end

	is_undefined : BOOLEAN is
			-- true if day, month and year are undefined, otherwise false
			do
				Result := not is_day_defined and not is_month_defined and not is_year_defined
			end



feature -- Setting Date value

	set_date(a_date_string: STRING) is
			-- Sets the Date values
			require
				string_not_void: a_date_string /= void
				string_length_correct: a_date_string.count = 10
			do
				m_date:=a_date_string
			end

	set_null_date is
			-- sets this instance to an undefined date
			do
				m_date := void
			end

feature -- Comparing

	is_newer_date (other: GEN_DATE): BOOLEAN is
			-- Is current newer then other
			require
				current_not_null: not Current.is_null_date
				other_not_null: not other.is_null_date
			local cur_date, other_date: INTEGER;
			do
				Result := false
			    if Current.is_year_defined and other.is_year_defined then
					cur_date := Current.year * 10000
					other_date := other.year * 10000
					if (Current.is_month_defined and other.is_month_defined) then
						cur_date := cur_date + Current.month * 100
						other_date := other_date + other.month * 100
						if (Current.is_day_defined and other.is_day_defined) then
							cur_date := cur_date + Current.day
							other_date := other_date + other.day
						end
					end
					Result := cur_date > other_date
			    end
			end

	is_equal_date (other: GEN_DATE): BOOLEAN is
			-- Is current date equal to other
			require
				current_not_null: not Current.is_null_date
				other_not_null: not other.is_null_date
			do
				if Current.is_full_defined and other.is_full_defined then
					Result:= Current.year=other.year and Current.month=other.month and Current.day=other.day
				else
					Result:=false
				end
			end

	is_newer_or_equal_date(other: GEN_DATE):BOOLEAN is
			-- true if current date is newer or equal other, otherwise false
			require
				current_not_null: not Current.is_null_date
				other_not_null: not other.is_null_date
			do
				if is_equal_date (other) then
					Result := true
				else
			        Result := is_newer_date (other)
				end
			end

invariant
	invariant_clause: True -- Your invariant here

end
