indexing
	description: "[
					  A genealogical database,
					  represented through a single, flat interface

					  Project 2 of the Software Engineering course 2008
					  ]"
	author: "Bernd Schoeller <bernd.schoeller@inf.ethz.ch>, Thomas Britschgi, Peter Bucher, Nicola Ranieri"
	date: "$Date: 2008-05-20 21:01:05 +0200 (Tue, 20 May 2008) $"
	revision: "$Revision: 14 $"

class
	GENEALOGY2

-- REMARKS:

-- Dates are represented as STRINGs. They use the form "TT.MM.YYYY",
-- where TT is the day using two digits, MM is the month using two
-- digits and YYYY is the year using four digits. ?? (year: ????) can
-- be used to represent unknown parts of the date. If the year is
-- unknown, the month needs to be unknown as well. If the month is
-- unknown, the day needs to be unknown as well.

-- Validity of all data entered should be sound, after every
-- operation. This must be enforced by the underlying implementation.
-- Your implementation must reject any operation that would make the
-- database unsound.

-- The data needs to be chronologically sound: a person cannot die
-- before he is born, or a father be younger than his child. It is not
-- necessary to model natural limits like an upper age limit or a
-- minimum or maximum age for child birth.

-- Marriages are not terminated by the death of one partner, but the
-- widow can marry another person without the need for a divorce -
-- this implicitly terminates the old marriage.

-- The whole interface implements "defensive programming", that means
-- that no extra preconditions should be added, but instead invalid
-- operations should report errors using the
-- 'last_operation_successful' and 'last_error' reporting
-- facilities. You are allowed (and encouraged) to add your
-- postconditions and invariants.


-- Rules:
--
-- all functions (procedures and queries) do set the operation_Successful flag except:
--		last_operation_successful
--		last_error
--		valid_date
--		number_of_registered_persons
--
--
-- add_male / add_female: 	This operation represents the birth of a person,
--							and the birth date is taken as event date and has
--							to be in the right chronological order of the events (def. prog.).
--							A person must have a unique positive ID (precond.),
--							a valid birth date (precond.) and a non-void (precond.), non-empty
--							first and last name (def. prog.).
--
--
-- register_death:	A person must exist to die (precond.) on a valid date (precond.).
--					The death date must be in right chronological order (def. prog.)
--					and must be still alive (def. prog.)
--
--
-- register_child:	Both parent and child must exist (precond.).
--					Since the registration of a child belongs to a birth,
--					the birth date of the child has to be in chronological order (def. prog.).
--					This guarantees, that the parent is at least as old as the child.
--					A child cannot be registered twice (def. prog.) and can only be registered
--					by one father and one mother (def. prog.), not two fathers or two mothers.
--					Both persons must be alive (def. prog.).
--					One cannot register a child by itself (def. prog.) and the child must not be an anchestor
--					of the parent, or there would be cycles (def. prog.). This is not captured by the chronological
--					order in case of equal birth dates.
--					If the parents birthdate is not defined, the child must be older than the grandparents,
--					but this is guaranteed by the global chronological order.
--					There are NO restriction about the relation between the parents (e.g. incest)
--
--					
--	adopt:	Parent and child must exist (precond.)
--			The date must be valid (precond.) and match ino the global chronological order (def. prog.).
--			The child must not be a child (adopted or natural) of parent (def. prog.).
--			The child must be younger than the parent (def. prog.). This is not handled by the chronological
--			order, since the event date is not the birthdate but the adoption date. If the date of the parent
--			is not comparable, than the check has to be extended on the grandparents,
--			than the great grandparents and so on.
--			Parent and child must be alive for adoption (def. prog.).
--			One cannot adopt itself (def. prog.) and the child must not be an anchestor
--			of the parent, or there would be cycles (def. prog.). This is not captured by the chronological
--			order in case of equal birth dates.
--			There are no restrictions on who can adopt (e.g. your grandfather, your sister etc.).
--			If a last name is set, it must not be empty.
--
--	
--	marry:	Both partners must exist (precond.) and marry on a valid date (precond.).
--			The marriage date must match into chronological order (def. prog.).
--			both partners must be alive (def. prog.), and not married to another living partner (def. prog.),
--			but may be widowed. One cannot marry himself (def. prog.), a direct sibling (def. prog.) or a direct ancestor (def. prog.).
--			Marriage between cousins, or divorced, not direct ancestors are allowed.
--			If last names are set, they have to be non-empty.
--
--	
--	divorce:	The person must exist (precond.) and married (def. prog.) to divorce on a valid date (precond.) which matches into
--				the global chronological order (def. prog.).
--				Although it is not necessary to divorce when widowed, the living partner (def. prog.) can decide to divorce.
--
--
-- Quality Assurance by: Design, contracts, test cases, mutation tests and coverage tests



create
	make_empty


feature {NONE} -- Initialization

	make_empty is
			-- Make an empty genealogical database.
		do
			create data.make (200) -- initial value, hash-table will resize itself if needed
			operation_Successful := true
			errorstring := ""
			create history.make
		end


feature -- Error reporting

	last_operation_successful: BOOLEAN is
			-- Was the last operation successful?
		do
			Result := operation_Successful
		end

	last_error: STRING is
			-- Reason for the last failed operation (single line string)
		require
			error_available: not last_operation_successful
		do
			Result := errorString
		ensure
			not_void: Result /= Void
		end


feature -- Individuals changes

	add_male (an_id: INTEGER; a_first_name, a_last_name, a_date: STRING) is
			-- Add a new male person with id `an_id', first name
			-- `a_first_name', last name `a_last_name' and birthday
			-- `a_date'.
		require
			id_positive: an_id > 0
			first_name_not_void: a_first_name /= Void
			last_name_not_void: a_last_name /= Void
			date_valid: valid_date (a_date)
			fresh_id: not has (an_id)
		local
			dummy : PERSON2
			date : DATE2
		do
			-- convert the string to a date
			create date.make( a_date )
			-- check for empty names
			if a_first_name.is_empty or a_last_name.is_empty then
				errorstring := "Names must not be empty"
				operation_successful := false
			-- check if the birth event matches into chronological order
			elseif not history.is_chronologically_correct(date) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- create and add the person
			else
				create dummy.make( an_id, true, a_first_name, a_last_name, date )
				data.extend( dummy, an_id )
				history.add_event (history.BIRTH, date)
				operation_successful := true
			end
		ensure
			data_extended: operation_successful implies data.count = old data.count + 1
		end

	add_female (an_id: INTEGER; a_first_name, a_last_name, a_date: STRING) is
			-- Add a new female person with id `an_id', first name
			-- `a_first_name', last name `a_last_name' and birthday
			-- `a_date'.
		require
			id_positive: an_id > 0
			first_name_not_void: a_first_name /= Void
			last_name_not_void: a_last_name /= Void
			date_valid: valid_date (a_date)
			fresh_id: not has (an_id)
		local
			dummy : PERSON2
			date : DATE2
		do
			-- convert the string to a date
			create date.make( a_date )
			-- check for empty names
			if a_first_name.is_empty or a_last_name.is_empty then
				errorstring := "Names must not be empty"
				operation_successful := false
			-- check if the birth event matches into chronological order
			elseif not history.is_chronologically_correct(date) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- create and add the person
			else
				create dummy.make( an_id, false, a_first_name, a_last_name, date )
				data.extend( dummy, an_id )
				history.add_event (history.BIRTH, date)
				operation_successful := true
			end
		ensure
			data_extended: operation_successful implies data.count = old data.count + 1
		end

	register_death (an_id: INTEGER; a_date: STRING) is
			-- Register the death of person with `an_id', dying on `a_date'.
		require
			date_valid: valid_date (a_date)
			valid_person: has (an_id)
		local
			person : PERSON2
			date : DATE2
		do
			-- get the person
			person := data.item( an_id )
			-- convert the string to a date
			create date.make( a_date )
			-- test if the death event matches into chronological order
			if not history.is_chronologically_correct(date) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- test if the person allready died
			elseif person.get_dateofdeath /= void then
				errorstring := "Couldn't assign date of death, because it's already set for this id"
				operation_successful := false
			-- set the death date
			else
				person.set_dateofdeath( date )
				history.add_event (history.DEATH, date)
				operation_successful := true
			end
		ensure
			date_not_void: operation_successful implies data.item (an_id ).get_dateofdeath /= void
		end


feature -- Individual status

	has (an_id: INTEGER): BOOLEAN is
			-- Is there a person with id `an_id'?
		do
			-- ID must be positive
			if an_id > 0 then
				Result := data.has( an_id )
				operation_successful := true
			else
				Result := false
				errorstring := "ID must bee greater than zero"
				operation_successful := false
			end
		ensure
			only_valid_ids_exist: (an_id <= 0) implies not Result
		end

	is_male (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' male?
		require
			valid_person: has (an_id)
		do
			Result := data.item( an_id ).is_male
			operation_successful := true
		ensure
			right_result: Result = data.item( an_id ).is_male
		end

	is_female (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' female?
		require
			valid_person: has (an_id)
		do
			Result := not data.item( an_id ).is_male
			operation_successful := true
		ensure
			right_result:Result = not data.item( an_id ).is_male
		end

	is_alive (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' alive?
		require
			valid_person: has (an_id)
		local
			person 	: PERSON2
		do
			person := data.item( an_id )
			if person.get_dateofdeath = void then
				Result := true
			else
				Result := false
			end
			operation_successful := true
		end

	is_dead (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' dead?
		require
			valid_person: has (an_id)
		do
			Result := not is_alive( an_id )
		end


feature -- Individual query

	first_name (an_id: INTEGER): STRING is
			-- First name of `an_id'
		require
			valid_person: has (an_id)
		do
			Result := data.item( an_id ).get_firstname
			operation_successful := true
		ensure
			correct_name: Result.is_equal( data.item( an_id ).get_firstname )
		end

	last_name (an_id: INTEGER): STRING is
			-- (Current) last name of `an_id'
		require
			valid_person: has (an_id)
		do
			Result := data.item( an_id ).get_lastname
			operation_successful := true
		ensure
			correct_name: Result.is_equal( data.item( an_id ).get_lastname )
		end

	birth_name (an_id: INTEGER): STRING is
			-- Birth name of `an_id'
		require
			valid_person: has (an_id)
		do
			Result := data.item( an_id ).get_birthname
			operation_successful := true
		ensure
			correct_name: Result.is_equal( data.item( an_id ).get_birthname )
		end

	birth_date (an_id: INTEGER): STRING is
			-- Birth date of `an_id'
		require
			valid_person: has (an_id)
		do
			Result := data.item( an_id ).get_dateofbirth.string_representation
			operation_successful := true
		ensure
			res_set: Result.is_equal( data.item( an_id ).get_dateofbirth.string_representation )
		end

	death_date (an_id: INTEGER): STRING is
			-- Death date of `an_id', Void if the person is not dead
		require
			valid_person: has (an_id)
		local
			date : DATE2
		do
			date := data.item( an_id ).get_dateofdeath
			if date /= Void then
				Result := date.string_representation
			else
				Result := Void
			end
			operation_successful := true
		ensure
			res_set: Result /= void implies Result.is_equal (data.item (an_id).get_dateofdeath.string_representation )
		end

	find (a_name: STRING): LIST [INTEGER] is
			-- Persons with first or last name equal to `a_name'
		require
			not_void: a_name /= Void
		do
			from
				data.start
				Result := create {LINKED_LIST[INTEGER]}.make
			until
				data.after
			loop
				if 	data.item_for_iteration.get_lastname.is_equal( a_name ) or
					data.item_for_iteration.get_firstname.is_equal( a_name ) then

					Result.extend ( data.key_for_iteration )
				end
				data.forth
			end
			operation_successful := true
		ensure
			not_void: Result /= Void
		end


feature -- Relations

	register_child (a_parent, a_child: INTEGER) is
			-- Register `a_child' as the (natural) child of `a_parent',
			-- making `a_parent' mother or father of `a_child'.
		require
			valid_parent: has (a_parent)
			valid_child: has (a_child)
		local
			parent	: PERSON2
			child	: PERSON2
		do
			-- initialize the parent
			parent := data.item( a_parent )
			--initialize the child
			child  := data.item( a_child )
			-- test if the birth event matches into chronological order
			if not history.is_chronologically_correct( child.get_dateofbirth ) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- is already a child (e.g. adoption)
			elseif parent.get_adoptedchildren.has( child ) or parent.get_children.has ( child ) then
				errorString := "Child is already a child of parent"
				operation_successful := false
			-- if one of them is allready dead
			elseif parent.get_dateofdeath /= Void or child.get_dateofdeath /= Void then
				errorString := "Parent or child is not alive"
				operation_successful := false
			-- cannot register the child by itself
			elseif a_parent = a_child then
				errorString := "Cannot be child of itself"
				operation_successful := false
			-- if child is ancestor of the parent (only possible if parent has the same age as the child)
			elseif is_ancestor( a_child, a_parent ) then
				errorString := "Child is ancestor of parent"
				operation_successful := false
			else
				-- case: father
				if parent.is_male then
					-- is father allready set
					if child.get_naturalfather /= void then
						errorString := "Natural father already set"
						operation_successful := false
					else
						child.set_naturalfather( parent )
						parent.addchild( child )
						history.add_event (history.CHILD_REGISTRATION, child.get_dateofbirth)
						operation_successful := true
					end
				-- case: mother
				else
					-- is mother allready set
					if child.get_naturalmother /= void then
						errorString := "Natural mother already set"
						operation_successful := false
					else
						child.set_naturalmother( parent )
						parent.addchild( child )
						history.add_event (history.CHILD_REGISTRATION, child.get_dateofbirth)
						operation_successful := true
					end
				end
			end
		end

	adopt (a_parent, a_child: INTEGER; a_last_name: STRING; a_date: STRING) is
			-- Adopt `a_child' by `a_parent' on `a_date', changing the
			-- last name of the child to `a_last_name'.
		require
			valid_parent: has (a_parent)
			valid_child: has (a_child)
			valid_date: valid_date (a_date)
		local
			parent : PERSON2
			child  : PERSON2
			date : DATE2
		do
			-- initialize the parent
			parent := data.item( a_parent )
			-- initialize the child
			child  := data.item( a_child  )
			-- convert the string to a date
			create date.make(a_date)
			-- test if the adopt event matches into chronological order
			if not history.is_chronologically_correct( date ) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- if it is already a child	
			elseif parent.get_adoptedchildren.has( child ) or parent.get_children.has( child ) then
				errorString := "Parent has allready this child"
				operation_successful := false
			-- child must be younger
			elseif not parent.is_earlier_born_than( child.get_dateofbirth ) then
				errorString := "Child must be younger than the parent"
				operation_successful := false
			-- both must be alive
			elseif parent.get_dateofdeath /= Void or child.get_dateofdeath /= Void then
				errorString := "Father and child must be alive for adoption"
				operation_successful := false
			-- cannot adopt itself
			elseif a_parent = a_child then
				errorString := "Cannot adopt itself"
				operation_successful := false
			-- if child is ancestor of the parent (only possible if parent has the same age as the child)
			elseif is_ancestor( a_child, a_parent ) then
				errorString := "Child is ancestor of parent"
				operation_successful := false
			-- if a last name is set but empty
			elseif a_last_name /= Void and then a_last_name.is_empty then
				errorstring := "If a last name is set, it must not be empty"
				operation_successful := false
			else
				-- case: father
				if parent.is_male then
					child.set_AdoptingFather( parent )
					if a_last_name /= Void then
						child.set_lastname (a_last_name)
					end
				-- case: mother
				else
					child.set_AdoptingMother( parent )
					if a_last_name /= Void then
						child.set_lastname (a_last_name)
					end
				end
				parent.adoptchild (child, date)
				history.add_event (history.ADOPTION, date)
				operation_successful := true
			end
		end


feature -- Relation status

	is_natural_father (a_parent, a_child: INTEGER): BOOLEAN is
			-- Is `a_parent' the natural father of `a_child'?
		require
			valid_parent: has (a_parent)
			valid_child: has (a_child)
		local
			parent : PERSON2
			child  : PERSON2
		do
			parent := data.item( a_parent )
			child  := data.item( a_child  )
			if parent.is_male then
				Result := child.get_naturalfather = parent
				operation_successful := true
			else
				Result := false
				errorstring := "Parent isn't even male..."
				operation_successful := false
			end
		ensure
			is_child_of: data.item(a_parent).get_children.has( data.item (a_child) ) = Result
		end

	is_natural_mother (a_parent, a_child: INTEGER): BOOLEAN is
			-- Is `a_parent' the natural mother of `a_child'?
		require
			valid_parent: has (a_parent)
			valid_child: has (a_child)
		local
			parent : PERSON2
			child  : PERSON2
		do
			parent := data.item( a_parent )
			child  := data.item( a_child  )
			if not parent.is_male then
				Result := child.get_naturalmother = parent
				operation_successful := true
			else
				Result := false
				errorstring := "Parent isn't even female..."
				operation_successful := false
			end
		ensure
			is_child_of: data.item(a_parent).get_children.has( data.item (a_child) ) = Result
		end

	is_parent (a_parent, a_child: INTEGER): BOOLEAN is
			-- Is `a_parent' a parent (natural or adopted) of `a_child'?
		require
			valid_parent: has (a_parent)
			valid_child: has (a_child)
		local
			parent : PERSON2
			child  : PERSON2
		do
			parent := data.item( a_parent )
			child  := data.item( a_child  )
			Result := child.get_parents.has( parent )
			operation_successful := true
		ensure
			is_child_of: data.item(a_parent).get_children.has( data.item (a_child) ) = Result or data.item(a_parent).get_adoptedchildren.has ( data.item (a_child) ) = Result
		end


feature -- Relation queries

	father (a_child: INTEGER): INTEGER is
			-- The (current) father of `a_child'
			-- 0 if unkown
		require
			valid_child: has (a_child)
		local
			papi : PERSON2
		do
			papi := data.item( a_child ).get_currentfather
			if papi /= void then
				Result := papi.get_id
				operation_successful := true
			else
				Result := 0
				errorString := "Father unknown"
				operation_successful := false
			end
		end

	mother (a_child: INTEGER): INTEGER is
			-- The (current) mother of `a_child'
			-- 0 if unknown
		require
			valid_child: has (a_child)
		local
			mami : PERSON2
		do
			mami := data.item( a_child ).get_currentmother
			if mami /= void then
				Result := mami.get_id
				operation_successful := true
			else
				Result := 0
				errorString := "Mother unknown"
				operation_successful := false
			end
		end

	parents (a_child: INTEGER): LIST [INTEGER] is
			-- List of all parents of `a_child', including adoptions
		require
			valid_child: has (a_child)
		do
			Result := data.item( a_child ).get_parent_ids
			operation_successful := true
		ensure
			not_void: Result /= Void
			father_included: (father(a_child) > 0) implies Result.has (father (a_child))
			mother_included: (mother(a_child) > 0) implies Result.has (mother (a_child))
		end

	children (a_parent: INTEGER): LIST [INTEGER] is
			-- All children (natural or adopted) of `a_parent'
		require
			valid_parent: has (a_parent)
		local
			parent : PERSON2
		do
			parent := data.item( a_parent )
			Result := parent.get_children_ids
			Result.append (parent.get_adoptedchildren_ids )
			operation_successful := true
		ensure
			not_void: Result /= Void
		end


feature -- Marriages

	marry (a_first_person,a_second_person: INTEGER;
			 a_first_last_name, a_second_last_name: STRING;
			 a_date: STRING) is
			-- `a_first_person' marries `a_second_person' on `a_date',
			-- resulting in the first person to change the last name to
			-- `a_first_last_name' (Void = no change) and the second
			-- person to change the last name to `a_second_last_name'
			-- (Void = no change).
		require
			valid_first: has (a_first_person)
			valid_second: has (a_second_person)
			valid_date: valid_date (a_date)
		local
			person : PERSON2
			partner : PERSON2
			tempPersonPartner : PERSON2
			tempPartnerPartner : PERSON2
			date : DATE2
		do
			-- initialize first person
			person := data.item ( a_first_person )
			-- initialize second person
			partner := data.item ( a_second_person )
			-- convert the string to s date
			create date.make( a_date )
			-- get current partners of the couple
			tempPersonPartner := person.get_currentpartner
			tempPartnerPartner := partner.get_currentpartner
			-- test if the birth event matches into chronological order
			if not history.is_chronologically_correct( date ) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- test if one is married and the spouse is still alive
			elseif is_married( a_first_person ) and then tempPersonPartner.get_dateofdeath = Void then
				errorString := "First person is still married"
				operation_successful := false
			elseif is_married( a_second_person ) and then tempPartnerPartner.get_dateofdeath = Void then
				errorString := "Second person is still married"
				operation_successful := false
			-- test if not allready dead
			elseif person.get_dateofdeath /= Void or partner.get_dateofdeath /= Void then
				errorString := "Persons must be alive to marry"
				operation_successful := false
			-- test for self marriage
			elseif a_first_person = a_second_person then
				errorString := "Cannot marry himself"
				operation_successful := false
			-- test for direct siblings (symmetric relation)
			elseif person.has_direct_sibling( partner ) then
				errorString := "Cannot marry a direct sibling"
				operation_successful := false
			-- test for direct ancestor in both directions
			elseif person.has_direct_ancestor( partner ) or partner.has_direct_ancestor( person ) then
				errorString := "Cannot marry a direct ancestor"
				operation_successful := false
			-- if last names are set, they have to be non-empty
			elseif (a_first_last_name /= Void and then a_first_last_name.is_empty) or (a_second_last_name /= Void and then a_second_last_name.is_empty) then
				errorstring := "If a last name is set, it must not be empty"
				operation_successful := false
			else
				person.marry (partner, a_first_last_name, date)
				partner.marry (person, a_second_last_name, date)
				history.add_event (history.MARRIAGE, date)
				operation_successful := true
			end
		end

	divorce (an_id: INTEGER; a_first_last_name, other_last_name: STRING;
				a_date: STRING) is
			-- `an_id' divorces from the current partner on `a_date',
			-- resulting in the person to change the last name to
			-- `a_last_name' (Void = no change) and the spouse to change
			-- the last name to `other_last_name' (Void = no change).
			-- divorcing does not require both partners to live
		require
			valid_id: has (an_id)
			valid_date (a_date)
		local
			person : PERSON2
			partner : PERSON2
			marriageDate : DATE2
			date : DATE2
		do
			-- initialize first person
			person := data.item (an_id)
			-- initialize second person
			partner := person.get_currentpartner
			-- convert the string to a date
			create date.make( a_date )
			-- get the marriage date
			if person.get_marriagedates.is_empty then
				marriageDate := Void
			else
				marriageDate := person.get_marriagedates.last
			end
			-- test if the birth event matches into chronological order
			if not history.is_chronologically_correct( date ) then
				errorstring := "Event does not match in chronological order"
				operation_successful := false
			-- test if married
			elseif not is_married( an_id ) or partner = Void or marriageDate = Void then
				errorString := "Person has to be married to divorce"
				operation_successful := false
			-- test if the initiator is alive
			elseif person.get_dateofdeath /= Void then
				errorString := "Person must be alive to divorce"
				operation_successful := false
			-- if last names are set, they have to be non-empty
			elseif (a_first_last_name /= Void and then a_first_last_name.is_empty) or (other_last_name /= Void and then other_last_name.is_empty) then
				errorstring := "If a last name is set, it must not be empty"
				operation_successful := false
			else
				person.divorce(a_first_last_name, date)
				partner.divorce(other_last_name, date)
				history.add_event (history.DIVORCE, date)
				operation_successful := true
			end
		end


feature -- Marriage status

	is_married (an_id: INTEGER): BOOLEAN is
			-- Is `an_id' currently married?
		require
			valid_id: has (an_id)
		local
			person : PERSON2
		do
			person := data.item ( an_id )
			Result := not (person.get_currentpartner = void)
			operation_successful := true
		end


feature -- Marriage queries

	partners (an_id: INTEGER): LIST [INTEGER] is
			-- Ordered list of all partners of `an_id'. Duplicates are
			-- possible if the same person was married multiple times.
		require
			valid_id: has (an_id)
		local
			person : PERSON2
		do
			person := data.item( an_id )
			Result := person.get_partner_ids
			operation_successful := true
		ensure
			not_void: Result /= Void
		end

	marriage_dates (an_id: INTEGER): LIST [STRING] is
			-- Date of all marriage dates of `an_id', corresponding to
			-- the `partners' query
		require
			valid_id: has (an_id)
		local
			person : PERSON2
			list : LIST[DATE2]
			resultList : LINKED_LIST[STRING]
		do
			create resultList.make
			person := data.item( an_id )
			list := person.get_marriagedates
			from
				list.start
			until
				list.after
			loop
				resultList.extend(list.item.string_representation)
				list.forth
			end
			Result := resultList
			operation_successful := true
		ensure
			not_void: Result /= Void
			same_length: Result.count = partners (an_id).count
		end

	divorce_dates (an_id: INTEGER): LIST [STRING] is
			-- Date of all divorce dates of `an_id', corresponding to the
			-- `partners' query; Void if the corresponding marriage was
			-- not divorced
		require
			valid_id: has (an_id)
		local
			person : PERSON2
			list : LIST[DATE2]
			resultList : LINKED_LIST[STRING]
		do
			create resultList.make
			person := data.item( an_id )
			list := person.get_divorcedates
			from
				list.start
			until
				list.after
			loop
--				Result.extend(list.item.string_representation)
				resultlist.extend(list.item.string_representation) -- fixed typo!
				list.forth
			end
			Result := resultList
			operation_successful := true
		ensure
			not_void: Result /= Void
			same_length: Result.count = partners (an_id).count
		end

feature -- Analysis status

	is_ancestor (a_first_id, a_second_id: INTEGER): BOOLEAN is
			-- Is `a_first_id' an ancestor of `a_second_id'?
			-- (including adoptions)
		require
			valid_first: has (a_first_id)
			valid_second: has (a_second_id)
		local
			descendant : PERSON2
			parentList : LIST[INTEGER]
		do
			descendant := data.item ( a_second_id )
			parentList := descendant.get_Parent_ids
			Result := false
			-- check all parents
			from
				parentList.start
			until
				parentList.after or
				Result = true
			loop
				-- if we have found the ancestor, we are finished
				if parentList.item = a_first_id then
					Result := true
					parentList.finish
				else
					-- recursion starts here, depth first search
					Result := is_ancestor(a_first_id, parentList.item)
				end
				parentList.forth
			end
			operation_successful := true
		end


feature -- Analysis queries

	closest_common_ancestor (a_first_id, a_second_id: INTEGER): INTEGER is
			-- The id of the closest common ancestor between `a_first_id'
			-- and `a_second_id'. If there is a common ancestor only
			-- following natural relationships, this one should be
			-- preferred.  Otherwise, the number of adoptions should be
			-- minimized. If there is more than one closest ancestor, the
			-- result is non-deterministic. 0 if there is none.
		require
			valid_first: has (a_first_id)
			valid_second: has (a_second_id)
		local
			ancestor_list_a : LINKED_LIST[INTEGER]
			ancestor_list_b : LINKED_LIST[INTEGER]
		do
			if a_first_id /= a_second_id then
				create ancestor_list_a.make
				create ancestor_list_b.make

				ancestor_list_a.extend (a_first_id)
				ancestor_list_b.extend (a_second_id)

				Result := cca_rec( create {TWO_WAY_LIST[INTEGER]}.make, create {TWO_WAY_LIST[INTEGER]}.make, ancestor_list_a, ancestor_list_b )
				operation_successful := true
			else
				errorstring := "first_id = second_id, one can't be his own ancestor"
				operation_successful := false
			end
		end


	shortest_relation (a_first_id, a_second_id: INTEGER): LIST [INTEGER] is
			-- The shortest relation between `a_first_id' and
			-- `a_second_id', following child/parent, adoption and (old
			-- and current) marriage relations. The first element in the
			-- list is `a_first_id', the last element in the list is
			-- `a_second_id'. Void if there is no relation between the
			-- two.
		require
			valid_first: has (a_first_id)
			valid_second: has (a_second_id)
		do
			if a_first_id /= a_second_id then

				Result := sr_dijkstra(a_first_id, a_second_id )
				operation_successful := true
			else
				-- TODO is this legal or not? if yes result: {a_first_id, a_second_id} or {a_first_id}  ?
				errorstring := "first_id = second_id, "
				operation_successful := false
			end
		ensure
			first_first_if_found: (Result /= Void) implies
										 Result.first = a_first_id
			second_last_if_found: (Result /= Void) implies
										 Result.last = a_second_id
		end


	best_relation (a_first_id, a_second_id: INTEGER): LIST [INTEGER] is
			-- The best relation between `a_first_id' and `a_second_id',
			-- following child/parent, adoption and marriage
			-- relations. The number of marriage relations should be
			-- minimized. For the same number of marriages, the number
			-- adoptions should be minimized. For the same number of
			-- adoptions, list size should be minimized. Void if there is
			-- no relation between the two.
		require
			valid_first: has (a_first_id)
			valid_second: has (a_second_id)
		local
			startList : LINKED_LIST[INTEGER]
			possible_best_relation : TUPLE[LINKED_LIST[INTEGER], INTEGER, INTEGER]
		do
			-- check whether a_first_id and a_second_id are different
			if a_first_id /= a_second_id then
				create startlist.make
				possible_best_relation := best_relation_recursion( [startlist, 0, 0], a_first_id, a_second_id )

				if possible_best_relation /= void then
					Result ?= possible_best_relation.item (1)
					Result.start
				else
					Result := void
				end

				operation_successful := true
			else
				errorstring := "first_id = second_id is not allowed in best_relation"
				operation_successful := false
			end
		ensure
			first_first_if_found: (Result /= Void) implies
										 Result.first = a_first_id
			second_last_if_found: (Result /= Void) implies
										 Result.last = a_second_id
		end


feature {NONE} -- Analysis queries private rountines

	cca_rec( ancestors_a, ancestors_b : TWO_WAY_LIST[INTEGER]; ancestors_to_add_a, ancestors_to_add_b : LIST[INTEGER] ) : INTEGER is
			-- recursion routine for cca
			-- ancestors_to_add contains initially only the two person ids which are to check.
		local
			new_ancestors_a 			: LINKED_LIST[INTEGER]
			new_NonNatural_Ancestors_a	: LINKED_LIST[INTEGER]
			new_ancestors_b 			: LINKED_LIST[INTEGER]
			new_NonNatural_Ancestors_b	: LINKED_LIST[INTEGER]
--			overlaps 				 : TWO_WAY_LIST[INTEGER]
			list					 	: LIST[INTEGER]
		do
			create new_ancestors_a.make
			create new_nonnatural_ancestors_a.make
			create new_ancestors_b.make
			create new_nonnatural_ancestors_b.make

--			overlaps := create {TWO_WAY_LIST[INTEGER]}.make
			Result := 0

			from
				ancestors_to_add_a.start
			until
				ancestors_to_add_a.after
			loop
				-- add the parents of the current ancestor_to_add_a to the new_ancestors list_ba. may be used later

				from
					list := data.item( ancestors_to_add_a.item ).get_naturalparent_ids
					list.start
				until
					list.after
				loop
					if not ancestors_a.has(list.item) then
						new_ancestors_a.extend(list.item)
					end
					list.forth
				end
				from
					list := data.item( ancestors_to_add_a.item ).get_parent_ids
					list.start
				until
					list.after
				loop
					if not ancestors_a.has(list.item) then
						new_nonnatural_ancestors_a.extend(list.item)
					end
					list.forth
				end

				insert_youngest_first ( ancestors_to_add_a.item, ancestors_a ) --
				ancestors_to_add_a.forth
			end


			from
				ancestors_to_add_b.start
			until
				ancestors_to_add_b.after
			loop
				-- add the parents of the current ancestor_to_add_b to the new_ancestors list_b. may be used later

				from
					list := data.item( ancestors_to_add_b.item ).get_naturalparent_ids
					list.start
				until
					list.after
				loop
					if not ancestors_b.has(list.item) then
						new_ancestors_b.extend(list.item)
					end
					list.forth
				end
				from
					list := data.item( ancestors_to_add_b.item ).get_parent_ids
					list.start
				until
					list.after
				loop
					if not ancestors_b.has(list.item) then
						new_nonnatural_ancestors_b.extend(list.item)
					end
					list.forth
				end

				insert_youngest_first ( ancestors_to_add_b.item, ancestors_b ) --
				ancestors_to_add_b.forth
			end

			-- ok now we've filled the ancestor_lists with all parents of the ancestors_to_add list given.
			-- in order, youngest first!
			-- is there an overlap?

			from
				ancestors_a.start
			until
				ancestors_a.after
			loop
				if ancestors_b.has (ancestors_a.item) then
					Result := ancestors_a.item  -- first overlap must be the youngest.
					ancestors_a.finish
				end
				ancestors_a.forth
			end


			if Result = 0 then
				-- no overlap found --> go deeper

				if not new_ancestors_a.is_empty or not new_ancestors_b.is_empty then
					Result := cca_rec( ancestors_a.deep_twin, ancestors_b.deep_twin, new_ancestors_a , new_ancestors_b ) -- make a deep_twin of ancestors, because otherwise we would have to remove
																			 -- the newly added ids of the next recursion step, if the recursion fails
				end

				if Result = 0 then
					-- no natural ancestors found, involve non natural, if there is anything more
					if not new_nonnatural_ancestors_a.is_empty or not new_nonnatural_ancestors_b.is_empty then
						Result := cca_rec( ancestors_a.deep_twin, ancestors_b.deep_twin, new_nonnatural_ancestors_a, new_nonnatural_ancestors_b )
					end
				end
			end
		end


	insert_youngest_first ( id : INTEGER; list : TWO_WAY_LIST[INTEGER] ) is
			-- inserts the id into the list in a way that the youngest id comes first!
		local
			bd : DATE2
			inserted : BOOLEAN
		do
			inserted := false
			bd := data.item (id).get_dateofbirth

			-- if list is empty
			if list.is_empty then
				-- just add id
				list.extend (id)
				inserted := true
			else
				-- check for correct position, before inserting
				from
					list.start
				until
					list.after
				loop
					if not data.item( list.item ).get_dateofbirth.is_later_than( bd )  then -- wenn list.item nicht spaeter als bd -->  bd ist juenger
						list.put_left (id)
						list.finish
						inserted := true
					end
					list.forth
				end
				if not inserted then
					list.extend (id)
					inserted := true
				end
			end
		 	check insertion_done: inserted end
		end


	sr_dijkstra( a_first_id, a_second_id : INTEGER ) : LINKED_LIST[INTEGER] is
			-- shortest relation recursion
		local
			distance 	: HASH_TABLE[ INTEGER, INTEGER ]
			previous 	: HASH_TABLE[ INTEGER, INTEGER ]
			notvisited  : LINKED_LIST[INTEGER]
			relatives 	: LINKED_LIST[INTEGER]
			person   	: PERSON2
			pathfound 	: BOOLEAN
			done 		: BOOLEAN
			personcount : INTEGER
			currentmin  : INTEGER
			current_id  : INTEGER
		do
--	 1. Create a distance list, a previous vertex list, a visited list, and a current vertex.				
			distance := create {HASH_TABLE[INTEGER, INTEGER]}.make (200)
			previous := create {HASH_TABLE[INTEGER, INTEGER]}.make (200)
			notvisited  := create {LINKED_LIST[INTEGER]}.make

			-- initialize previous with information that a_first_id has no previous id
			previous.force ( 0, a_first_id )

--   2. All the values in the distance list are set to infinity except the starting vertex which is set to zero.
--   3. All values in visited list are set to false.
			from
				personcount := data.count
				data.start
			until
				data.after
			loop
				distance.force (personcount, data.key_for_iteration ) -- set distance bigger than max
				notvisited.extend (data.key_for_iteration)
				data.forth
			end
			distance.force (0, a_first_id)

			from
				pathfound := false
				done := false
			until
				notvisited.is_empty or
				pathfound = true or
				done = true
			loop

--  5. Find the vertex with smallest path so far
				from
					currentmin := personcount
					person := void
					notvisited.start
				until
					notvisited.after
				loop
					if distance.item (notvisited.item) < currentmin then
					   person := data.item ( notvisited.item )
					   currentmin := distance.item (notvisited.item)
					end
					notvisited.forth
				end

				if person /= void then  -- if person is still void, we haven't found a person with < personcount --> all reachable checked
--   6. Mark the current vertex as visited.					
					notvisited.start
					notvisited.search (person.get_id)
					notvisited.remove
--   7. Update distance and previous lists based on those vertices which can be immediately reached from the current vertex.			
					from
						relatives := create {LINKED_LIST[INTEGER]}.make
						relatives.append( person.get_parent_ids )
						relatives.append( person.get_partner_ids )
						relatives.append( person.get_children_ids )
						relatives.append( person.get_adoptedchildren_ids )
						relatives.start
					until
						relatives.after
					loop
						if distance.item (person.get_id) + 1 < distance.item(relatives.item) then
							distance.force ( distance.item (person.get_id) + 1, relatives.item )
							previous.force ( person.get_id, relatives.item )
						end

						if relatives.item = a_second_id then
							pathfound := true
							relatives.finish
						end
						relatives.forth
					end
				else
					done := true
				end
			end
-- move along the 'previous' table to find the path
			if pathfound then
				Result := create {LINKED_LIST[INTEGER]}.make
				from
					current_id := a_second_id
				until
					not previous.has (current_id)
				loop
					Result.put_front (current_id)
					current_id := previous.item (current_id)
				end
			else
				Result := void
			end
		end


	best_relation_recursion( masterList : TUPLE[LINKED_LIST[INTEGER], INTEGER, INTEGER]; current_id, a_destination_id : INTEGER ) : TUPLE[LINKED_LIST[INTEGER], INTEGER, INTEGER] is
			-- finds all paths from one person id to another by following child/parent, adoption and marriage relations
			-- and avoiding cycles. Then return the one with the fewest marriage relations.
			-- if there are equals, return the one with the fewest adoptions
			-- if there are equals, return the shortest one.
			-- returns void, if there exists no relation.
		local
			countMarriage, countAdoptions : INTEGER
			working_list                  : LINKED_LIST[INTEGER]
			curr_recursion_list,
			master_recursion_list         : LINKED_LIST[INTEGER]
			new_masterList                : TUPLE[LINKED_LIST[INTEGER], INTEGER, INTEGER]
			person                        : PERSON2
		do
			-- set variables
			person := data.item (current_id)
			countMarriage := masterList.integer_item (2)
			countAdoptions := masterList.integer_item (3)

			master_recursion_list ?= masterList.item(1)
			create curr_recursion_list.make
			curr_recursion_list.copy (master_recursion_list);

			Result := void

			-- if destination is reached, abort recursion
			if (current_id = a_destination_id) then
				curr_recursion_list.extend (current_id)
				result := [curr_recursion_list, countmarriage, countadoptions];
			elseif curr_recursion_list.has (current_id) then
				-- if current_id already exists in the recursion list, abort recursion to avoid cycles
				result := void
			else -- this id is visited the first time

				-- add current_id to the recursion_list
				curr_recursion_list.extend (current_id)

				create working_list.make

				-- check whether there is a best relation through the natural children
				from
					working_list.wipe_out
					working_list.append( person.get_children_ids )
					working_list.start
				until
					working_list.after
				loop
					new_masterlist := best_relation_recursion ([curr_recursion_list, countmarriage, countadoptions], working_list.item, a_destination_id)
					Result := getBestMasterList(new_masterlist, Result, a_destination_id)
					working_list.forth
				end

				-- check whether there is a best relation through the parents
				from
					working_list.wipe_out
					working_list.append( person.get_parent_ids )
					working_list.start
				until
					working_list.after
				loop
					-- here it has to be ckeched whether current_id was adopted by his parent
					-- if yes, we have to increase the adoption count by one!
					if data.item(working_list.item).get_adoptedchildren_ids.has(current_id) then
						-- current_id was adopted, countadoptions is increased on recursion
						new_masterlist := best_relation_recursion ([curr_recursion_list, countmarriage, countadoptions+1], working_list.item, a_destination_id)
						Result := getBestMasterList(new_masterlist, Result, a_destination_id)
					elseif data.item(working_list.item).get_children_ids.has(current_id) then
						-- current_id is natural child
						new_masterlist := best_relation_recursion ([curr_recursion_list, countmarriage, countadoptions], working_list.item, a_destination_id)
						Result := getBestMasterList(new_masterlist, Result, a_destination_id)
					else
						check not_adopted_and_not_natural_child: false end
					end

					working_list.forth
				end

				-- check whether there is a best relation through adopted children
				from
					working_list.wipe_out
					working_list.append( person.get_adoptedchildren_ids )
					working_list.start
				until
					working_list.after
				loop
					new_masterlist := best_relation_recursion ([curr_recursion_list, countmarriage, countadoptions+1], working_list.item, a_destination_id)
					Result := getBestMasterList(new_masterlist, Result, a_destination_id)
					working_list.forth
				end

				-- check whether there is a best relation through the partner
				if person.get_currentpartner /= void then
					new_masterlist := best_relation_recursion ([curr_recursion_list, countmarriage+1, countadoptions], person.get_currentpartner.get_id, a_destination_id)
					Result := getBestMasterList(new_masterlist, Result, a_destination_id)
				end
			end
		end


feature -- public for testing purposes

	getBestMasterList(m1, m2 : TUPLE[LINKED_LIST[INTEGER], INTEGER, INTEGER]; destination_id : INTEGER) : TUPLE[LINKED_LIST[INTEGER], INTEGER, INTEGER] is
			-- returns the better of the two masterlists acoording to the following criteria:
				-- if one list is void, the other is returned.
				-- if both lists are void, void is returned.

				-- if one list does not reach destination_id, the oter is returned.
				-- if both lists don't reach destination_id, void is returned.

				-- if both lists reach destination_id:
					-- return the one with the fewest marriage relations.
					-- if there are equals, return the one with the fewest adoptions.
					-- if there are equals, return the shortest one.
					-- if there are equals, return an arbitrary one.
		local
			rec_list1, rec_list2 : LINKED_LIST[INTEGER]
		do
			Result := void
			if m1 = void then
				Result := m2
			end
			if m2 = void then
				Result := m1
			end

			if (m1 /= void) and (m2 /= void) then
				-- both masterLists are not void
				rec_list1 ?= m1.item (1)
				rec_list2 ?= m2.item (1)
				if rec_list1.last = destination_id and rec_list2.last /= destination_id  then
					Result := m1
				elseif rec_list1.last /= destination_id and rec_list2.last = destination_id then
					Result := m2
				elseif rec_list1.last /= destination_id and rec_list2.last /= destination_id then
					Result := void
				else
					if m1.integer_item(2) < m2.integer_item(2) then
						-- m1 contains less marriage relations than m2
						Result := m1
					elseif m1.integer_item(2) > m2.integer_item(2) then
						-- m2 contains less marriage relations than m1
						Result := m2
					else
						if m1.integer_item(3) < m2.integer_item(3) then
							-- m1 contains less adoption relations than m2
							Result := m1
						elseif m1.integer_item(3) > m2.integer_item(3) then
							-- m2 contains less adoption relations than m1
							Result := m2
						else
							if rec_list1.count < rec_list2.count then
								-- m1 contains less nodes than m2
								Result := m1
							elseif rec_list1.count > rec_list2.count then
								-- m2 contains less nodes than m1
								Result := m2
							else
								-- Both paths are equally good and equally long,
								-- so the result can be an arbitrary one of the two
								-- we pick the first one
								Result := m1
							end
						end
					end
					 -- both masterlists have a complete path
				end
			end
		end


feature -- Date validation

	valid_date (a_date: STRING): BOOLEAN is
			-- Is `a_date' a valid date?
			-- a_date must not be empty, but we cannot add an require
			-- hence, this has to be checked by defensive programming
		local
			date : DATE2
		do
			create date.make (a_date)
			Result := date.is_valid
		ensure
			void_implies_false: (a_date = Void) implies not Result
			wrong_length_implies_false: (a_date.count /= 10) implies not Result
		end


feature --test suporter

	number_of_registered_persons : INTEGER is
		-- returns the number of items in 'data'
		do
			Result := data.count
		end



feature {NONE} -- Implementation

	data 				   : HASH_TABLE [PERSON2, INTEGER]
	errorString			   : STRING
	operation_Successful   : BOOLEAN

	shortest_relation_list : LIST[INTEGER]


	history : HISTORY
		-- the history, containing all events in chronological order

end
