indexing
	description: "Objects that represent a chronological order on dates"
	author: "N.Ranieri"
	date: "$Today$"
	revision: "$1.0$"

class
	HISTORY

create
	make

feature -- Initialisation

	make is
		-- initialises the history
		do
			-- create an array containing the event strings for possible prints
			create event.make(1,6)
			event[1] := "Birth"
			event[2] := "Death"
			event[3] := "Child registration"
			event[4] := "Adoption"
			event[5] := "Marriage"
			event[6] := "Divorce"

			-- initialize the history
			create history.make
		end

feature -- event handling

	is_chronologically_correct(a_date : DATE2) : BOOLEAN is
		-- checks if the date is later, equal or not comparable to the last event
		-- new events can only be added, when this function returns true
		-- all events use the date passed as argument, except child registration, which uses the birth date
		require
			valid_date: a_date.is_valid
			date_not_void: a_date /= void
		local
			found : BOOLEAN
			currentDate : DATE2
		do
			-- special case: empty history returns allways true
			if history.is_empty then
				Result := true
				found := true
			else
				found := false
			end

			-- loop throu the history until you can compare a date
			-- return true if before is reached withaout a result,
			-- or if it is the latest event
			from
				history.finish
				Result := true
			until
				history.before or found
			loop
				-- down cast
				currentDate ?= history.item.item(2)
				-- is_later_than only guarantees success when returning true ...
				if a_date.is_later_than( currentDate ) then
					Result := true
					found := true
				-- ... hence check the other way around too ..
				elseif currentDate.is_later_than( a_date ) then
					Result := false
					found := true
				end
				-- ... or it is not comparable
				history.back
			end
		end

	add_event(type : INTEGER; a_date : DATE2) is
		-- appends an event to the history
		require
			valid_date: a_date.is_valid
			chronological_order: is_chronologically_correct( a_date )
			valid_type: type > 0 and type < 7
			date_not_void: a_date /= void
		local
			tuple : TUPLE[INTEGER,DATE2]
		do
			create tuple.default_create
			tuple.put (type, 1)
			tuple.put (a_date, 2)
			history.extend (tuple)
		ensure
			inserted: old history.count + 1 = history.count
		end

feature --constants

	BIRTH 				: INTEGER is 1
	DEATH 				: INTEGER is 2
	CHILD_REGISTRATION 	: INTEGER is 3
	ADOPTION 			: INTEGER is 4
	MARRIAGE 			: INTEGER is 5
	DIVORCE 			: INTEGER is 6

feature {NONE} -- Implementation

	history : LINKED_LIST[TUPLE[INTEGER,DATE2]]
	event : ARRAY[STRING]

end
