indexing
	description: "Pseudo-random sequnce with some additional features"
	author: ""
	date: "$Date$"
	revision: "$Revision$"

class
	MY_RANDOM

inherit
	RANDOM
		rename
			make as default_make
		end

	DOUBLE_COMPARISON
create make

feature -- Initialization
	make is
			-- Initialize structure using time-dependent seed
		local
			t: TIME
			r: DOUBLE
		do
			create t.make_now
			r := t.fine_seconds
			set_seed((default_seed*(r - r.floor_real_64)).floor)
			start
		end

feature -- Access
	bounded_item (min, max: INTEGER): INTEGER is
			-- Random integer from [min..max]
		require
			bounds_valid: max >= min
		local
			rmin, rmax: REAL
		do
			rmin := min
			rmax := max
			Result := (real_item * (rmax - rmin + 1) + rmin).floor
			if Result > max then Result := max
			end
		ensure
			result_not_too_small: Result >= min
			result_not_too_large: Result <= max
		end

	bounded_nonzero_item (min, max: INTEGER): INTEGER is
			-- Random integer from [min..max], but not 0
		require
			bounds_valid: max >= min
			not_just_zero: max = min implies max /= 0
		do
			from
			until
				Result /= 0
			loop
				Result := bounded_item (min, max)
				forth
			end
		ensure
			non_zero: Result /= 0
		end

	bounded_real_item (min, max: DOUBLE): DOUBLE is
			-- Random real from [min..max]
		require
			bounds_valid: max >= min
		do
			Result := min + (max - min)*real_item
		ensure
			result_not_too_small: Result >= min
			result_not_too_large: Result <= max
		end

	alpha_string_item (n: INTEGER): STRING is
			-- Random string of `n' characters
		require
			n_not_negative: n >= 0
		local
			i: INTEGER
		do
			create Result.make (n)
			from
				i := 1
			until
				i > n
			loop
				if decide (0.5) then
					forth
					Result.append_character (bounded_item (('A').code, ('Z').code).to_character_8)
					forth
				else
					forth
					Result.append_character (bounded_item (('a').code, ('z').code).to_character_8)
					forth
				end
				i := i + 1
			end
		ensure
			result_exists: Result /= Void
			count_correct: Result.count = n
		end

	alpha_string_le_item (n: INTEGER): STRING is
			-- Random string of not more than `n' characters
		require
			n_not_negative: n >= 0
		local
			k: INTEGER
		do
			k := bounded_item (0, n)
			forth
			Result := alpha_string_item (k)
		ensure
			result_exists: Result /= Void
		end

	alpha_string_nonempty_item (n: INTEGER): STRING is
			-- Random string of not more than `n' characters, but not empty
		require
			n_positive: n > 0
		local
			k: INTEGER
		do
			k := bounded_item (1, n)
			forth
			Result := alpha_string_item (k)
		ensure
			result_exists: Result /= Void
		end

	void_probability: REAL is 0.1

	alpha_string_or_void_item (n: INTEGER): STRING is
			-- Random string of not more than `n' characters or Void
		require
			n_not_negative: n >= 0
		do
			if not decide (void_probability) then
				forth
				Result := alpha_string_le_item (n)
			end
			forth
		end

	decide (p: DOUBLE): BOOLEAN is
			-- True with probability `p', False otherwise
		require
			p_is_probability: (p >= 0.0) and then (p <= 1.0)
		do
			Result := real_item < p
		end

	sample (distribution: ARRAY [DOUBLE]): INTEGER is
			-- Random index from [`distribution.lower', `distribution.upper'] with probability defined by `distribution'
		require
			distribution_exists: distribution /= Void
			distribution_valid: is_distribution_valid (distribution)
		local
			sum: DOUBLE
		do
			from
				Result := distribution.lower
				sum := bounded_real_item (0.0, 1.0)
			until
				sum <= 0.0
			loop
				check Result <= distribution.upper end -- Because sum in [0, 1] and distribution.totak in [0, 1]
				sum := sum - distribution[Result]
				Result := Result + 1
			end

			if Result > distribution.lower then
				Result := Result - 1
			end
		ensure
			index_not_too_small: Result >= distribution.lower
			index_not_too_large: Result <= distribution.upper
		end

	is_distribution_valid (distribution: ARRAY [DOUBLE]): BOOLEAN is
			-- Is `distribution' a valid discrete distribution?
		require
			distribution_exists: distribution /= Void
		local
			i: INTEGER
			has_negative: BOOLEAN
			sum: DOUBLE
		do
			from
				i := distribution.lower
			until
				i > distribution.upper
			loop
				if distribution[i] < 0.0 then
					has_negative := True
					i := distribution.upper
				else
					sum := sum + distribution [i]
				end
				i := i + 1
			end
			Result := not has_negative and approx_equal (sum, 1.0)
		end

	permutation (n: INTEGER): ARRAY [INTEGER] is
			-- Random permutation of numbers [1..n]
		require
			n_positive: n > 0
		local
			list: LINKED_LIST [INTEGER]
			i: INTEGER
		do
			create Result.make (1, n)
			create list.make
			from
				i := 1
			until
				i > n
			loop
				list.extend (i)
				i := i + 1
			end
			from
			until
				list.is_empty
			loop
				i := bounded_item (1, list.count)
				forth
				list.go_i_th (i)
				Result.put (list.item, n - list.count + 1)
				list.remove
			end
		end
end
